'''
Functions for manage auth and credentials
'''
# pylint: disable=redefined-outer-name, protected-access, no-name-in-module, import-error,line-too-long
from functools import lru_cache

import sys

from google.cloud import bigquery, storage

import pydata_google_auth


from basedosdados.upload.base import Base

SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
]


def reauth():
    '''
    Reauth user credentials
    '''

    pydata_google_auth.get_user_credentials(
        SCOPES, credentials_cache=pydata_google_auth.cache.REAUTH
    )


def credentials(from_file=False, reauth=False):
    '''
    Get user credentials
    '''

    #check if is running in colab
    if "google.colab" in sys.modules:
        from google.colab import auth  # pylint: disable=import-outside-toplevel

        auth.authenticate_user()
        return None

    if from_file:
        return Base()._load_credentials(mode="prod")

    if reauth:
        return pydata_google_auth.get_user_credentials(
            SCOPES, credentials_cache=pydata_google_auth.cache.REAUTH
        )

    return pydata_google_auth.get_user_credentials(
            SCOPES,
    )



@lru_cache(256)
def google_client(billing_project_id, from_file, reauth):
    '''
    Get Google Cloud client for bigquery and storage
    '''

    return dict(
        bigquery=bigquery.Client(
            credentials=credentials(from_file=from_file, reauth=reauth),
            project=billing_project_id,
        ),
        storage=storage.Client(
            credentials=credentials(from_file=from_file, reauth=reauth),
            project=billing_project_id,
        ),
    )



"""
CLI package for the application.
"""

# pylint: disable=locally-disabled, multiple-statements, fixme, line-too-long, too-many-arguments, invalid-name, too-many-lines, protected-access, unused-argument, no-value-for-parameter, redefined-builtin

import click
from basedosdados.upload.base import Base
from basedosdados.upload.dataset import Dataset
from basedosdados.upload.table import Table
from basedosdados.upload.storage import Storage
from basedosdados.upload.metadata import Metadata

import basedosdados as bd
from basedosdados.exceptions import BaseDosDadosException
from ckanapi import CKANAPIError


@click.group()
@click.option("--templates", default=None, help="Templates path")
@click.option("--bucket_name", default=None, help="Project bucket name")
@click.option("--metadata_path", default=None, help="Folder to store metadata")
@click.version_option(package_name="basedosdados")
@click.pass_context
def cli(ctx, templates, bucket_name, metadata_path):
    """
    Function to define the CLI.
    """

    ctx.obj = dict(
        templates=templates,
        bucket_name=bucket_name,
        metadata_path=metadata_path,
    )


@click.group(name="dataset")
@click.pass_context
def cli_dataset(ctx):
    """
    Command to manage datasets.
    """


@cli_dataset.command(name="init", help="Initialize metadata files of dataset")
@click.argument("dataset_id")
@click.option(
    "--replace",
    is_flag=True,
    help="Whether to replace current metadata files",
)
@click.pass_context
def init_dataset(ctx, dataset_id, replace):
    """
    Initialize metadata files of dataset.
    """

    d = Dataset(dataset_id=dataset_id, **ctx.obj).init(replace=replace)

    click.echo(
        click.style(
            f"Dataset `{dataset_id}` folder and metadata were created at {d.metadata_path}",
            fg="green",
        )
    )


def mode_text(mode, verb, obj_id):
    """
    Returns a specific text for each mode.
    """

    if mode == "all":
        text = f"Datasets `{obj_id}` and `{obj_id}_staging` were {verb} in BigQuery"
    elif mode == "staging":
        text = f"Dataset `{obj_id}_stating` was {verb} in BigQuery"
    elif mode == "prod":
        text = f"Dataset `{obj_id}` was {verb} in BigQuery"

    return text


@cli_dataset.command(name="create", help="Create dataset on BigQuery")
@click.argument("dataset_id")
@click.option(
    "--mode", "-m", default="all", help="What datasets to create [prod|staging|all]"
)
@click.option(
    "--if_exists",
    default="raise",
    help="[raise|update|replace|pass] if dataset alread exists",
)
@click.option(
    "--dataset_is_public",
    default=True,
    help="Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.",
)
@click.option(
    "--location",
    default=None,
    help="Location of dataset data. List of possible region names locations: https://cloud.google.com/bigquery/docs/locations",
)
@click.pass_context
def create_dataset(ctx, dataset_id, mode, if_exists, dataset_is_public, location):
    """
    Create dataset on BigQuery.
    """

    Dataset(dataset_id=dataset_id, **ctx.obj).create(
        mode=mode,
        if_exists=if_exists,
        dataset_is_public=dataset_is_public,
        location=location,
    )

    click.echo(
        click.style(
            mode_text(mode, "created", dataset_id),
            fg="green",
        )
    )


@cli_dataset.command(name="update", help="Update dataset on BigQuery")
@click.argument("dataset_id")
@click.option(
    "--mode", "-m", default="all", help="What datasets to create [prod|staging|all]"
)
@click.pass_context
def update_dataset(ctx, dataset_id, mode, location):
    """
    Update dataset on BigQuery.
    """

    Dataset(dataset_id=dataset_id, **ctx.obj).update(mode=mode, location=location)

    click.echo(
        click.style(
            mode_text(mode, "updated", dataset_id),
            fg="green",
        )
    )


@cli_dataset.command(name="publicize", help="Make a dataset public")
@click.argument("dataset_id")
@click.option(
    "--dataset_is_public",
    default=True,
    help="Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.",
)
@click.pass_context
def publicize_dataset(ctx, dataset_id, dataset_is_public):
    """
    Publicize dataset.
    """

    Dataset(dataset_id=dataset_id, **ctx.obj).publicize(
        dataset_is_public=dataset_is_public
    )

    click.echo(
        click.style(
            f"Dataset `{dataset_id}` became public!",
            fg="green",
        )
    )


@cli_dataset.command(name="delete", help="Delete dataset")
@click.argument("dataset_id")
@click.option(
    "--mode", "-m", default="all", help="What datasets to create [prod|staging|all]"
)
@click.pass_context
def delete_dataset(ctx, dataset_id, mode):
    """
    Delete dataset.
    """

    if click.confirm(f"Are you sure you want to delete `{dataset_id}`?"):

        Dataset(dataset_id=dataset_id, **ctx.obj).delete(mode=mode)

    click.echo(
        click.style(
            mode_text(mode, "deleted", dataset_id),
            fg="green",
        )
    )


@click.group(name="table")
def cli_table():
    """
    Command to manage tables.
    """


@cli_table.command(name="init", help="Create metadata files")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--data_sample_path",
    default=None,
    help="Sample data used to pre-fill metadata",
    type=click.Path(exists=True),
)
@click.option(
    "--if_folder_exists",
    default="raise",
    help="[raise|replace|pass] actions if table folder exists",
)
@click.option(
    "--if_table_config_exists",
    default="raise",
    help="[raise|replace|pass] actions if table config files already exist",
)
@click.option(
    "--source_format",
    default="csv",
    help="Data source format. Only 'csv' is supported. Defaults to 'csv'.",
)
@click.option(
    "--force_columns",
    default=False,
    help="Overwrite columns with local columns.",
)
@click.option(
    "--columns_config_url_or_path",
    default=None,
    help="google sheets URL. Must be in the format https://docs.google.com/spreadsheets/d/<table_key>/edit#gid=<table_gid>. The sheet must contain the column name: 'coluna' and column description: 'descricao'.",
)
@click.pass_context
def init_table(
    ctx,
    dataset_id,
    table_id,
    data_sample_path,
    if_folder_exists,
    if_table_config_exists,
    source_format,
    force_columns,
    columns_config_url_or_path,
):
    """
    Initialize table metadata.
    """

    t = Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).init(
        data_sample_path=data_sample_path,
        if_folder_exists=if_folder_exists,
        if_table_config_exists=if_table_config_exists,
        source_format=source_format,
        force_columns=force_columns,
        columns_config_url_or_path=columns_config_url_or_path,
    )

    click.echo(
        click.style(
            f"Table `{table_id}` folder and metadata were created at {t.metadata_path}{dataset_id}",
            fg="green",
        )
    )


@cli_table.command(name="create", help="Create stagging table in BigQuery")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--path",
    "-p",
    type=click.Path(exists=True),
    default=None,
    help="Path of data folder or file.",
)
@click.option(
    "--if_table_exists",
    default="raise",
    help="[raise|replace|pass] actions if table exists",
)
@click.option(
    "--force_dataset",
    default=True,
    help="Whether to automatically create the dataset folders and in BigQuery",
)
@click.option(
    "--if_storage_data_exists",
    default="raise",
    help="[raise|replace|pass] actions if table data already exists at Storage",
)
@click.option(
    "--if_table_config_exists",
    default="raise",
    help="[raise|replace|pass] actions if table config files already exist",
)
@click.option(
    "--source_format",
    default="csv",
    help="Data source format. Only 'csv' is supported. Defaults to 'csv'.",
)
@click.option(
    "--force_columns",
    default=False,
    help="Overwrite columns with local columns.",
)
@click.option(
    "--columns_config_url_or_path",
    default=None,
    help="Path to the local architeture file or a public google sheets URL. Path only suports csv, xls, xlsx, xlsm, xlsb, odf, ods, odt formats. Google sheets URL must be in the format https://docs.google.com/spreadsheets/d/<table_key>/edit#gid=<table_gid>.",
)
@click.option(
    "--dataset_is_public",
    default=True,
    help="Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.",
)
@click.option(
    "--location",
    default=None,
    help="Location of dataset data. List of possible region names locations: https://cloud.google.com/bigquery/docs/locations",
)
@click.option(
    "--chunk_size",
    default=None,
    help="The size of a chunk of data whenever iterating (in bytes). This must be a multiple of 256 KB per the API specification.",
)
@click.pass_context
def create_table(
    ctx,
    dataset_id,
    table_id,
    path,
    if_table_exists,
    force_dataset,
    if_storage_data_exists,
    if_table_config_exists,
    source_format,
    force_columns,
    columns_config_url_or_path,
    dataset_is_public,
    location,
    chunk_size,
):

    """
    Create staging table in BigQuery.
    """

    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).create(
        path=path,
        if_table_exists=if_table_exists,
        force_dataset=force_dataset,
        if_storage_data_exists=if_storage_data_exists,
        if_table_config_exists=if_table_config_exists,
        source_format=source_format,
        force_columns=force_columns,
        columns_config_url_or_path=columns_config_url_or_path,
        dataset_is_public=dataset_is_public,
        location=location,
        chunk_size=chunk_size,
    )

    click.echo(
        click.style(
            f"Table `{dataset_id}_staging.{table_id}` was created in BigQuery",
            fg="green",
        )
    )


@cli_table.command(name="update", help="Update tables in BigQuery")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--mode",
    default="all",
    help="Choose a table from a dataset to update [prod|staging|all]",
)
@click.pass_context
def update_table(ctx, dataset_id, table_id, mode):
    """
    Update tables in BigQuery.
    """

    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).update(
        mode=mode,
    )

    click.echo(
        click.style(
            f"All tables `{dataset_id}*.{table_id}` were updated in BigQuery",
            fg="green",
        )
    )


@cli_table.command(
    name="update_columns", help="Update columns fields in tables_config.yaml "
)
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--columns_config_url_or_path",
    default=None,
    help="""\nFills columns in table_config.yaml automatically using a public google sheets URL or a local file. Also regenerate
        \npublish.sql and autofill type using bigquery_type.\n

    \nThe sheet must contain the columns:\n
        - name: column name\n
        - description: column description\n
        - bigquery_type: column bigquery type\n
        - measurement_unit: column mesurement unit\n
        - covered_by_dictionary: column related dictionary\n
        - directory_column: column related directory in the format <dataset_id>.<table_id>:<column_name>\n
        - temporal_coverage: column temporal coverage\n
        - has_sensitive_data: the column has sensitive data\n
        - observations: column observations\n
    \nArgs:\n
    \ncolumns_config_url_or_path (str): Path to the local architeture file or a public google sheets URL.\n
        Path only suports csv, xls, xlsx, xlsm, xlsb, odf, ods, odt formats.\n
        Google sheets URL must be in the format https://docs.google.com/spreadsheets/d/<table_key>/edit#gid=<table_gid>.\n
""",
)
@click.pass_context
def update_columns(ctx, dataset_id, table_id, columns_config_url_or_path):
    """
    Update columns fields in tables_config.yaml
    """

    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).update_columns(
        columns_config_url_or_path=columns_config_url_or_path,
    )

    click.echo(
        click.style(
            f"Columns from `{dataset_id}.{table_id}` were updated in table_config.yaml",
            fg="green",
        )
    )


@cli_table.command(name="publish", help="Publish staging table to prod")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--if_exists",
    default="raise",
    help="[raise|replace] actions if table exists",
)
@click.pass_context
def publish_table(ctx, dataset_id, table_id, if_exists):
    """
    Publish staging table to prod.
    """

    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).publish(
        if_exists=if_exists,
    )

    click.echo(
        click.style(
            f"Table `{dataset_id}.{table_id}` was published in BigQuery",
            fg="green",
        )
    )


@cli_table.command(name="delete", help="Delete BigQuery table")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option("--mode", help="Which table to delete [prod|staging]", required=True)
@click.pass_context
def delete_table(ctx, dataset_id, table_id, mode):
    """
    Delete BigQuery table.
    """

    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).delete(
        mode=mode,
    )


@cli_table.command(name="append", help="Append new data to existing table")
@click.argument("dataset_id")
@click.argument("table_id")
@click.argument("filepath", type=click.Path(exists=True))
@click.option("--partitions", help="Data partition as `value=key/value2=key2`")
@click.option(
    "--if_exists",
    default="raise",
    help="[raise|replace|pass] if file alread exists",
)
@click.option(
    "--chunk_size",
    default=None,
    help="The size of a chunk of data whenever iterating (in bytes). This must be a multiple of 256 KB per the API specification.",
)
@click.pass_context
def upload_table(
    ctx, dataset_id, table_id, filepath, partitions, if_exists, chunk_size
):

    """
    Upload data to BigQuery table.
    """
    Table(table_id=table_id, dataset_id=dataset_id, **ctx.obj).append(
        filepath=filepath,
        partitions=partitions,
        if_exists=if_exists,
        chunk_size=chunk_size,
    )

    click.echo(
        click.style(
            f"Data was added to `{dataset_id}.{table_id}`",
            fg="green",
        )
    )


@click.group(name="storage")
def cli_storage():
    """
    Commands for Google Cloud Storage.
    """


@cli_storage.command(name="init", help="Create bucket and initial folders")
@click.option("--bucket_name", default="basedosdados", help="Bucket name")
@click.option(
    "--replace",
    is_flag=True,
    help="Whether to replace current bucket files",
)
@click.option(
    "--very-sure/--not-sure",
    default=False,
    help="Are you sure that you want to replace current bucket files?",
)
@click.pass_context
def init_storage(ctx, bucket_name, replace, very_sure):
    """
    Initialize bucket and folders.
    """

    # TODO: Create config file to store bucket_name, etc...
    ctx.obj.pop("bucket_name")
    Storage(bucket_name=bucket_name, **ctx.obj).init(
        replace=replace, very_sure=very_sure
    )

    click.echo(
        click.style(
            f"Bucket `{bucket_name}` was created",
            fg="green",
        )
    )


@cli_storage.command(name="upload", help="Upload file to bucket")
@click.argument("dataset_id")
@click.argument("table_id")
@click.argument("filepath", type=click.Path(exists=True))
@click.option(
    "--mode",
    "-m",
    required=True,
    help="Where to save the file [raw|staging|header|auxiliary_files|architecture|all]",
)
@click.option("--partitions", help="Data partition as `value=key/value2=key2`")
@click.option(
    "--if_exists",
    default="raise",
    help="[raise|replace|pass] if file alread exists",
)
@click.option(
    "--chunk_size",
    default=None,
    help="The size of a chunk of data whenever iterating (in bytes). This must be a multiple of 256 KB per the API specification.",
)
@click.pass_context
def upload_storage(
    ctx, dataset_id, table_id, filepath, mode, partitions, if_exists, chunk_size
):

    '''
    Upload file to bucket.
    '''

    ctx.obj.pop("bucket_name")
    Storage(dataset_id, table_id, **ctx.obj).upload(
        filepath=filepath,
        mode=mode,
        partitions=partitions,
        if_exists=if_exists,
        chunk_size=chunk_size,
    )

    click.echo(
        click.style(
            "Data was added",
            fg="green",
        )
    )


@cli_storage.command(name="download", help="Download file from bucket")
@click.argument("dataset_id")
@click.argument("table_id")
@click.argument("savepath", type=click.Path(exists=True))
@click.option(
    "--filename",
    "-f",
    default="*",
    help="filename to download single file. If * downloads all files from bucket folder",
)
@click.option(
    "--mode",
    "-m",
    default="raw",
    help="Where to download data from [raw|staging|header|auxiliary_files|architecture]",
)
@click.option("--partitions", help="Data partition as `value=key/value2=key2`")
@click.option(
    "--if_not_exists",
    default="raise",
    help="[raise|pass] if file file not found at bucket folder",
)
@click.pass_context
def download_storage(
    ctx, dataset_id, table_id, filename, savepath, partitions, mode, if_not_exists
):
    """
    Download file from bucket.
    """
    Storage(dataset_id, table_id, **ctx.obj).download(
        filename, savepath, partitions, mode, if_not_exists
    )

    click.echo(
        click.style(
            f"Data was downloaded to `{savepath}`",
            fg="green",
        )
    )


@cli_storage.command(name="delete_table", help="Delete table from bucket")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--mode",
    "-m",
    required=True,
    default="staging",
    help="Where to delete the file from [raw|staging|header|auxiliary_files|architecture]",
)
@click.option(
    "--bucket_name",
    default=None,
    help="Bucket from which to delete data, you can change it to delete from a bucket other than yours",
)
@click.option("--not_found_ok", default=False, help="what to do if table not found")
@click.pass_context
def storage_delete_table(ctx, dataset_id, table_id, mode, not_found_ok, bucket_name):
    """
    Delete table from bucket.
    """
    Storage(dataset_id, table_id, **ctx.obj).delete_table(
        mode=mode, not_found_ok=not_found_ok, bucket_name=bucket_name
    )
    click.echo(
        click.style(
            f"Data was deleted from bucket `{bucket_name}`",
            fg="green",
        )
    )


@cli_storage.command(name="copy_table", help="Copy table to your bucket")
@click.argument("dataset_id")
@click.argument("table_id")
@click.option("--source_bucket_name", required=True, default="basedosdados")
@click.option(
    "--dst_bucket_name",
    default=None,
    help="Bucket where data will be copied to, defaults to your bucket",
)
@click.option(
    "--mode",
    "-m",
    default="staging",
    help="which bucket folder to get the table [raw|staging|header|auxiliary_files|architecture]",
)
@click.pass_context
def storage_copy_table(
    ctx, dataset_id, table_id, source_bucket_name, dst_bucket_name, mode
):
    """
    Copy table from another bucket.
    """
    Storage(dataset_id, table_id, **ctx.obj).copy_table(
        source_bucket_name=source_bucket_name,
        destination_bucket_name=dst_bucket_name,
        mode=mode,
    )


@click.group(name="list")
def cli_list():
    """
    CLI list commands.
    """


@cli_list.command(name="datasets", help="List datasets available at given project_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.option(
    "--filter_by",
    default=None,
    help="Filter your search, must be a string",
)
@click.option(
    "--with_description",
    default=False,
    help="[bool]Fetch short description for each dataset",
)
@click.pass_context
def cli_list_datasets(ctx, project_id, filter_by, with_description):
    """
    List datasets available at given project_id
    """
    bd.list_datasets(
        query_project_id=project_id,
        filter_by=filter_by,
        with_description=with_description,
    )


@cli_list.command(name="dataset_tables", help="List tables available at given dataset")
@click.argument("dataset_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.option(
    "--filter_by",
    default=None,
    help="Filter your search, must be a string",
)
@click.option(
    "--with_description",
    default=False,
    help="[bool]Fetch short description for each table",
)
@click.pass_context
def cli_list_dataset_tables(ctx, dataset_id, project_id, filter_by, with_description):
    """
    List tables available at given dataset.
    """
    bd.list_dataset_tables(
        dataset_id=dataset_id,
        query_project_id=project_id,
        filter_by=filter_by,
        with_description=with_description,
    )


@click.group(name="get")
def cli_get():
    """
    Get commands.
    """


@cli_get.command(
    name="dataset_description", help="Get the full description for given dataset"
)
@click.argument("dataset_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_dataset_description(ctx, dataset_id, project_id):
    """
    Get the full description for given dataset
    """
    bd.get_dataset_description(
        dataset_id=dataset_id,
        query_project_id=project_id,
    )


@cli_get.command(
    name="table_description", help="Get the full description for given table"
)
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_table_description(ctx, dataset_id, table_id, project_id):
    """
    Get the full description for given table
    """
    bd.get_table_description(
        dataset_id=dataset_id,
        table_id=table_id,
        query_project_id=project_id,
    )


@cli_get.command(
    name="table_columns",
    help="Get fields names,types and description for columns at given table",
)
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_table_columns(
    ctx,
    dataset_id,
    table_id,
    project_id,
):
    """
    Get fields names,types and description for columns at given table
    """
    bd.get_table_columns(
        dataset_id=dataset_id,
        table_id=table_id,
        query_project_id=project_id,
    )


@click.group(name="metadata")
def cli_metadata():
    """
    CLI metadata commands.
    """


@cli_metadata.command(name="create", help="Creates new metadata config file")
@click.argument("dataset_id")
@click.argument("table_id", required=False)
@click.option(
    "--if_exists",
    default="raise",
    help="[raise|replace|pass] if metadata config file alread exists",
)
@click.option(
    "--columns",
    default=[],
    help="Data columns. Example: --columns=col1,col2",
    callback=lambda _, __, x: x.split(",") if x else [],
)
@click.option(
    "--partition_columns",
    default=[],
    help="Columns that partition the data. Example: --partition_columns=col1,col2",
    callback=lambda _, __, x: x.split(",") if x else [],
)
@click.option(
    "--force_columns",
    default=False,
    help="Overwrite columns with local columns.",
)
@click.option(
    "--table_only",
    default=True,
    help=(
        "Force the creation of `table_config.yaml` file only if `dataset_conf"
        "ig.yaml` doesn't exist."
    ),
)
@click.pass_context
def cli_create_metadata(
    ctx,
    dataset_id,
    table_id,
    if_exists,
    columns,
    partition_columns,
    force_columns,
    table_only,
):
    """
    Creates new metadata config file.
    """

    m = Metadata(dataset_id, table_id, **ctx.obj).create(
        if_exists=if_exists,
        columns=columns,
        partition_columns=partition_columns,
        force_columns=force_columns,
        table_only=table_only,
    )

    click.echo(
        click.style(
            f"Metadata file was created at `{m.filepath}`",
            fg="green",
        )
    )


@cli_metadata.command(
    name="is_updated", help="Check if user's local metadata is updated"
)
@click.argument("dataset_id")
@click.argument("table_id", required=False)
@click.pass_context
def cli_is_updated_metadata(ctx, dataset_id, table_id):
    """
    Check if user's local metadata is updated.
    """
    m = Metadata(dataset_id, table_id, **ctx.obj)

    if m.is_updated():
        msg, color = "Local metadata is updated.", "green"
    else:
        msg = (
            "Local metadata is out of date. Please run `basedosdados metadata"
            " create` with the flag `if_exists=replace` to get the updated da"
            "ta."
        )
        color = "red"

    click.echo(click.style(msg, fg=color))


@cli_metadata.command(name="validate", help="Validate user's local metadata")
@click.argument("dataset_id")
@click.argument("table_id", required=False)
@click.pass_context
def cli_validate_metadata(ctx, dataset_id, table_id):
    """
    Validate user's local metadata.
    """
    m = Metadata(dataset_id, table_id, **ctx.obj)

    try:
        m.validate()
        msg, color = "Local metadata is valid.", "green"
    except BaseDosDadosException as e:
        msg = (
            f"Local metadata is invalid. Please check the traceback below for"
            f" more information on how to fix it:\n\n{repr(e)}"
        )
        color = "red"

    click.echo(click.style(msg, fg=color))


@cli_metadata.command(name="publish", help="Publish user's local metadata")
@click.argument("dataset_id")
@click.argument("table_id", required=False)
@click.option(
    "--all",
    default=False,
    help=(
        "Force the publishment of metadata specified in both `dataset_config."
        "yaml` and `table_config.yaml` at once."
    ),
)
@click.option(
    "--if_exists",
    default="raise",
    help=("Define what to do in case metadata already exists in CKAN."),
)
@click.option(
    "--update_locally",
    default=False,
    help=("Update local metadata with the new CKAN metadata on publish."),
)
@click.pass_context
def cli_publish_metadata(
    ctx,
    dataset_id,
    table_id,
    all,
    if_exists,
    update_locally,
):
    """
    Publish user's local metadata.
    """
    m = Metadata(dataset_id, table_id, **ctx.obj)

    try:
        m.publish(all=all, if_exists=if_exists, update_locally=update_locally)
        msg, color = "Local metadata has been published.", "green"
    except (CKANAPIError, BaseDosDadosException, AssertionError) as e:
        msg = (
            f"Local metadata couldn't be published due to an error. Pleas"
            f"e check the traceback below for more information on how to "
            f"fix it:\n\n{repr(e)}"
        )
        color = "red"

    click.echo(click.style(msg, fg=color))


@click.group(name="config")
def cli_config():
    """
    CLI config commands.
    """


@cli_config.command(name="init", help="Initialize configuration")
@click.option(
    "--overwrite",
    default=False,
    help="Wheteher to overwrite current config",
)
@click.pass_context
def init(ctx, overwrite):
    """
    Initialize configuration.
    """

    Base(overwrite_cli_config=overwrite, **ctx.obj)


@cli_config.command(name="refresh_template", help="Overwrite current templates")
@click.pass_context
def init_refresh_templates(ctx):
    """
    Initialize configuration.
    """

    Base(**ctx.obj)._refresh_templates()


# Allow anomalous backslash in string: '\ ' (it's used in gcloud sdk)
# pylint: disable=W1401
@click.command(
    name="download",
    help=(
        "Downloads data do SAVEPATH. SAVEPATH must point to a .csv file.\n\n"
        "Example: \n\n"
        'basedosdados download data.csv  --query="select * from basedosdados.br_ibge_pib.municipio limit 10" \ --billing_project_id=basedosdados-dev'
    ),
)
@click.argument(
    "savepath",
    type=click.Path(exists=False),
)
@click.option(
    "--query",
    default=None,
    help="A SQL Standard query to download data from BigQuery",
)
@click.option(
    "--dataset_id",
    default=None,
    help="Dataset_id, enter with table_id to download table",
)
@click.option(
    "--table_id",
    default=None,
    help="Table_id, enter with dataset_id to download table ",
)
@click.option(
    "--query_project_id",
    default=None,
    help="Which project the table lives. You can change this you want to query different projects.",
)
@click.option(
    "--billing_project_id",
    default=None,
    help="Project that will be billed. Find your Project ID here https://console.cloud.google.com/projectselector2/home/dashboard",
)
@click.option(
    "--limit",
    default=None,
    help="Number of rows returned",
)
@click.pass_context
def cli_download(
    ctx,
    savepath,
    query,
    dataset_id,
    table_id,
    query_project_id,
    billing_project_id,
    limit,
):
    """
    Download data from BigQuery.
    """

    bd.download(
        savepath=savepath,
        dataset_id=dataset_id,
        table_id=table_id,
        query=query,
        query_project_id=query_project_id,
        billing_project_id=billing_project_id,
        limit=limit,
    )

    click.echo(
        click.style(
            f"Table was downloaded to `{savepath}`",
            fg="green",
        )
    )


@click.command(
    name="reauth",
    help="Reauthorize credentials.",
)
def cli_reauth():
    """
    Reauthorize credentials.
    """

    bd.reauth()


cli.add_command(cli_dataset)
cli.add_command(cli_table)
cli.add_command(cli_storage)
cli.add_command(cli_config)
cli.add_command(cli_download)
cli.add_command(cli_reauth)
cli.add_command(cli_list)
cli.add_command(cli_get)
cli.add_command(cli_metadata)


if __name__ == "__main__":

    cli()



'''
Class for managing the files in cloud storage.
'''
# pylint: disable=invalid-name, too-many-arguments, undefined-loop-variable,line-too-long,broad-except,R0801
import time
from pathlib import Path
import sys
import traceback

from tqdm import tqdm
from loguru import logger

from basedosdados.upload.base import Base
from basedosdados.exceptions import BaseDosDadosException

# google retryble exceptions. References: https://googleapis.dev/python/storage/latest/retry_timeout.html#module-google.cloud.storage.retry


class Storage(Base):
    """
    Manage files on Google Cloud Storage.
    """

    def __init__(self, dataset_id, table_id, **kwargs):
        super().__init__(**kwargs)

        self.bucket = self.client["storage_staging"].bucket(self.bucket_name)
        self.dataset_id = dataset_id.replace("-", "_")
        self.table_id = table_id.replace("-", "_")

    @staticmethod
    def _resolve_partitions(partitions):

        if isinstance(partitions, dict):

            return "/".join(f"{k}={v}" for k, v in partitions.items()) + "/"

        if isinstance(partitions, str):

            if partitions.endswith("/"):
                partitions = partitions[:-1]

            # If there is no partition
            if len(partitions) == 0:
                return ""

            # It should fail if there is folder which is not a partition
            try:
                # check if it fits rule
                {b.split("=")[0]: b.split("=")[1] for b in partitions.split("/")}
            except IndexError as e:
                raise Exception(f"The path {partitions} is not a valid partition") from e

            return partitions + "/"

        raise Exception(f"Partitions format or type not accepted: {partitions}")

    def _build_blob_name(self, filename, mode, partitions=None):
        '''
        Builds the blob name.
        '''

        # table folder
        blob_name = f"{mode}/{self.dataset_id}/{self.table_id}/"

        # add partition folder
        if partitions is not None:

            blob_name += self._resolve_partitions(partitions)

        # add file name
        blob_name += filename

        return blob_name

    def init(self, replace=False, very_sure=False):
        """Initializes bucket and folders.

        Folder should be:

        * `raw` : that contains really raw data
        * `staging` : preprocessed data ready to upload to BigQuery

        Args:
            replace (bool): Optional.
                Whether to replace if bucket already exists
            very_sure (bool): Optional.
                Are you aware that everything is going to be erased if you
                replace the bucket?

        Raises:
            Warning: very_sure argument is still False.
        """

        if replace:
            if not very_sure:
                raise Warning(
                    "\n********************************************************"
                    "\nYou are trying to replace all the data that you have "
                    f"in bucket {self.bucket_name}.\nAre you sure?\n"
                    "If yes, add the flag --very_sure\n"
                    "********************************************************"
                )
            self.bucket.delete(force=True)

        self.client["storage_staging"].create_bucket(self.bucket)

        for folder in ["staging/", "raw/"]:

            self.bucket.blob(folder).upload_from_string("")

    def upload(
        self,
        path,
        mode="all",
        partitions=None,
        if_exists="raise",
        chunk_size=None,
        **upload_args,
    ):
        """Upload to storage at `<bucket_name>/<mode>/<dataset_id>/<table_id>`. You can:

        * Add a single **file** setting `path = <file_path>`.

        * Add a **folder** with multiple files setting `path =
          <folder_path>`. *The folder should just contain the files and
          no folders.*

        * Add **partitioned files** setting `path = <folder_path>`.
          This folder must follow the hive partitioning scheme i.e.
          `<table_id>/<key>=<value>/<key2>=<value2>/<partition>.csv`
          (ex: `mytable/country=brasil/year=2020/mypart.csv`).

        *Remember all files must follow a single schema.* Otherwise, things
        might fail in the future.

        There are 6 modes:

        * `raw` : should contain raw files from datasource
        * `staging` : should contain pre-treated files ready to upload to BiqQuery
        * `header`: should contain the header of the tables
        * `auxiliary_files`: should contain auxiliary files from eache table
        * `architecture`: should contain the architecture sheet of the tables
        * `all`: if no treatment is needed, use `all`.

        Args:
            path (str or pathlib.PosixPath): Where to find the file or
                folder that you want to upload to storage

            mode (str): Folder of which dataset to update [raw|staging|header|auxiliary_files|architecture|all]

            partitions (str, pathlib.PosixPath, or dict): Optional.
                *If adding a single file*, use this to add it to a specific partition.

                * str : `<key>=<value>/<key2>=<value2>`
                * dict: `dict(key=value, key2=value2)`

            if_exists (str): Optional.
                What to do if data exists

                * 'raise' : Raises Conflict exception
                * 'replace' : Replace table
                * 'pass' : Do nothing
            chunk_size (int): Optional
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
                If not specified, the chunk_size of the blob itself is used. If that is not specified, a default value of 40 MB is used.

            upload_args ():
                Extra arguments accepted by [`google.cloud.storage.blob.Blob.upload_from_file`](https://googleapis.dev/python/storage/latest/blobs.html?highlight=upload_from_filename#google.cloud.storage.blob.Blob.upload_from_filename)
        """

        if (self.dataset_id is None) or (self.table_id is None):
            raise Exception("You need to pass dataset_id and table_id")

        path = Path(path)

        if path.is_dir():
            paths = [
                f
                for f in path.glob("**/*")
                if f.is_file() and f.suffix in [".csv", ".parquet", "parquet.gzip"]
            ]

            parts = [
                (
                    filepath.as_posix()
                    .replace(path.as_posix() + "/", "")
                    .replace(str(filepath.name), "")
                )
                for filepath in paths
            ]

        else:
            paths = [path]
            parts = [partitions or None]

        self._check_mode(mode)

        mode = (
            ["raw", "staging", "header", "auxiliary_files", "architecture"]
            if mode == "all"
            else [mode]
        )
        for m in mode:

            for filepath, part in tqdm(list(zip(paths, parts)), desc="Uploading files"):

                blob_name = self._build_blob_name(filepath.name, m, part)

                blob = self.bucket.blob(blob_name, chunk_size=chunk_size)

                if not blob.exists() or if_exists == "replace":

                    upload_args["timeout"] = upload_args.get("timeout", None)

                    blob.upload_from_filename(str(filepath), **upload_args)

                elif if_exists == "pass":

                    pass

                else:
                    raise BaseDosDadosException(
                        f"Data already exists at {self.bucket_name}/{blob_name}. "
                        "If you are using Storage.upload then set if_exists to "
                        "'replace' to overwrite data \n"
                        "If you are using Table.create then set if_storage_data_exists "
                        "to 'replace' to overwrite data."
                    )

        logger.success(
            " {object} {filename}_{mode} was {action}!",
            filename=filepath.name,
            mode=mode,
            object="File",
            action="uploaded",
        )

    def download(
        self,
        filename="*",
        savepath=".",
        partitions=None,
        mode="raw",
        if_not_exists="raise",
    ):

        """Download files from Google Storage from path `mode`/`dataset_id`/`table_id`/`partitions`/`filename` and replicate folder hierarchy
        on save,

        There are 5 modes:
        * `raw` : should contain raw files from datasource
        * `staging` : should contain pre-treated files ready to upload to BiqQuery
        * `header`: should contain the header of the tables
        * `auxiliary_files`: should contain auxiliary files from eache table
        * `architecture`: should contain the architecture sheet of the tables

        You can also use the `partitions` argument to choose files from a partition

        Args:
            filename (str): Optional
                Specify which file to download. If "*" , downloads all files within the bucket folder. Defaults to "*".

            savepath (str):
                Where you want to save the data on your computer. Must be a path to a directory.

            partitions (str, dict): Optional
                If downloading a single file, use this to specify the partition path from which to download.

                * str : `<key>=<value>/<key2>=<value2>`
                * dict: `dict(key=value, key2=value2)`


            mode (str): Optional
                Folder of which dataset to update.[raw|staging|header|auxiliary_files|architecture]

            if_not_exists (str): Optional.
                What to do if data not found.

                * 'raise' : Raises FileNotFoundError.
                * 'pass' : Do nothing and exit the function

        Raises:
            FileNotFoundError: If the given path `<mode>/<dataset_id>/<table_id>/<partitions>/<filename>` could not be found or there are no files to download.
        """

        # Prefix to locate files within the bucket
        prefix = f"{mode}/{self.dataset_id}/{self.table_id}/"

        # Add specific partition to search prefix
        if partitions:
            prefix += self._resolve_partitions(partitions)

        # if no filename is passed, list all blobs within a given table
        if filename != "*":
            prefix += filename

        blob_list = list(self.bucket.list_blobs(prefix=prefix))

        # if there are no blobs matching the search raise FileNotFoundError or return
        if not blob_list:
            if if_not_exists == "raise":
                raise FileNotFoundError(f"Could not locate files at {prefix}")
            return

        # download all blobs matching the search to given savepath
        for blob in tqdm(blob_list, desc="Download Blob"):

            # parse blob.name and get the csv file name
            csv_name = blob.name.split("/")[-1]

            # build folder path replicating storage hierarchy
            blob_folder = blob.name.replace(csv_name, "")

            # replicate folder hierarchy
            (Path(savepath) / blob_folder).mkdir(parents=True, exist_ok=True)

            # download blob to savepath
            savepath = f"{savepath}/{blob.name}"
            blob.download_to_filename(filename=savepath)

        logger.success(
            " {object} {object_id}_{mode} was {action} at: {path}!",
            object_id=self.dataset_id,
            mode=mode,
            object="File",
            action="downloaded",
            path={str(savepath)}
        )

    def delete_file(self, filename, mode, partitions=None, not_found_ok=False):
        """Deletes file from path `<bucket_name>/<mode>/<dataset_id>/<table_id>/<partitions>/<filename>`.

        Args:
            filename (str): Name of the file to be deleted

            mode (str): Folder of which dataset to update [raw|staging|header|auxiliary_files|architecture|all]

            partitions (str, pathlib.PosixPath, or dict): Optional.
                Hive structured partition as a string or dict

                * str : `<key>=<value>/<key2>=<value2>`
                * dict: `dict(key=value, key2=value2)`

            not_found_ok (bool): Optional.
                What to do if file not found
        """

        self._check_mode(mode)

        mode = (
            ["raw", "staging", "header", "auxiliary_files", "architecture"]
            if mode == "all"
            else [mode]
        )

        for m in mode:

            blob = self.bucket.blob(self._build_blob_name(filename, m, partitions))

            if blob.exists() or not blob.exists() and not not_found_ok:
                blob.delete()
            else:
                return

        logger.success(
            " {object} {filename}_{mode} was {action}!",
            filename=filename,
            mode=mode,
            object="File",
            action="deleted",
        )

    def delete_table(self, mode="staging", bucket_name=None, not_found_ok=False):
        """Deletes a table from storage, sends request in batches.

        Args:
            mode (str): Folder of which dataset to update [raw|staging|header|auxiliary_files|architecture]
                Folder of which dataset to update. Defaults to "staging".

            bucket_name (str):
                The bucket name from which to delete the table. If None, defaults to the bucket initialized when instantiating the Storage object.
                (You can check it with the Storage().bucket property)

            not_found_ok (bool): Optional.
                What to do if table not found

        """

        prefix = f"{mode}/{self.dataset_id}/{self.table_id}/"

        if bucket_name is not None:

            table_blobs = list(
                self.client["storage_staging"]
                .bucket(f"{bucket_name}")
                .list_blobs(prefix=prefix)
            )

        else:

            table_blobs = list(self.bucket.list_blobs(prefix=prefix))

        if not table_blobs:
            if not_found_ok:
                return
            raise FileNotFoundError(
                f"Could not find the requested table {self.dataset_id}.{self.table_id}"
            )
        # Divides table_blobs list for maximum batch request size
        table_blobs_chunks = [
            table_blobs[i : i + 999] for i in range(0, len(table_blobs), 999)
        ]

        for i, source_table in enumerate(
            tqdm(table_blobs_chunks, desc="Delete Table Chunk")
        ):
            counter = 0
            while counter < 10:
                try:
                    with self.client["storage_staging"].batch():
                        for blob in source_table:
                            blob.delete()
                    break
                except Exception:
                    print(
                        f"Delete Table Chunk {i} | Attempt {counter}: delete operation starts again in 5 seconds...",
                    )
                    time.sleep(5)
                    counter += 1
                    traceback.print_exc(file=sys.stderr)
        logger.success(
            " {object} {object_id}_{mode} was {action}!",
            object_id=self.table_id,
            mode=mode,
            object="Table",
            action="deleted",
        )

    def copy_table(
        self,
        source_bucket_name="basedosdados",
        destination_bucket_name=None,
        mode="staging",
    ):
        """Copies table from a source bucket to your bucket, sends request in batches.

        Args:
            source_bucket_name (str):
                The bucket name from which to copy data. You can change it
                to copy from other external bucket.

            destination_bucket_name (str): Optional
                The bucket name where data will be copied to.
                If None, defaults to the bucket initialized when instantiating the Storage object (You can check it with the
                Storage().bucket property)

            mode (str): Folder of which dataset to update [raw|staging|header|auxiliary_files|architecture]
                Folder of which dataset to update. Defaults to "staging".
        """

        source_table_ref = list(
            self.client["storage_staging"]
            .bucket(source_bucket_name)
            .list_blobs(prefix=f"{mode}/{self.dataset_id}/{self.table_id}/")
        )

        if not source_table_ref:
            raise FileNotFoundError(
                f"Could not find the requested table {self.dataset_id}.{self.table_id}"
            )

        if destination_bucket_name is None:

            destination_bucket = self.bucket

        else:

            destination_bucket = self.client["storage_staging"].bucket(
                destination_bucket_name
            )

        # Divides source_table_ref list for maximum batch request size
        source_table_ref_chunks = [
            source_table_ref[i : i + 999] for i in range(0, len(source_table_ref), 999)
        ]

        for i, source_table in enumerate(
            tqdm(source_table_ref_chunks, desc="Copy Table Chunk")
        ):
            counter = 0
            while counter < 10:
                try:
                    with self.client["storage_staging"].batch():
                        for blob in source_table:
                            self.bucket.copy_blob(
                                blob,
                                destination_bucket=destination_bucket,
                            )
                    break
                except Exception:
                    print(
                        f"Copy Table Chunk {i} | Attempt {counter}: copy operation starts again in 5 seconds...",
                    )
                    counter += 1
                    time.sleep(5)
                    traceback.print_exc(file=sys.stderr)
        logger.success(
            " {object} {object_id}_{mode} was {action}!",
            object_id=self.table_id,
            mode=mode,
            object="Table",
            action="copied",
        )



"""
Module for manage dataset to the server.
"""
# pylint: disable=line-too-long, fixme, invalid-name,line-too-long,unnecessary-lambda-assignment
from pathlib import Path
from loguru import logger

from google.cloud import bigquery
from google.api_core.exceptions import Conflict

from basedosdados.upload.base import Base
from basedosdados.upload.metadata import Metadata


class Dataset(Base):
    """
    Manage datasets in BigQuery.
    """

    def __init__(self, dataset_id, **kwargs):
        super().__init__(**kwargs)

        self.dataset_id = dataset_id.replace("-", "_")
        self.dataset_folder = Path(self.metadata_path / self.dataset_id)
        self.metadata = Metadata(self.dataset_id, **kwargs)

    @property
    def dataset_config(self):
        """
        Dataset config file.
        """

        return self._load_yaml(
            self.metadata_path / self.dataset_id / "dataset_config.yaml"
        )

    def _loop_modes(self, mode="all"):
        """
        Loop modes.
        """

        mode = ["prod", "staging"] if mode == "all" else [mode]
        dataset_tag = lambda m: f"_{m}" if m == "staging" else ""

        return (
            {
                "client": self.client[f"bigquery_{m}"],
                "id": f"{self.client[f'bigquery_{m}'].project}.{self.dataset_id}{dataset_tag(m)}",
            }
            for m in mode
        )

    @staticmethod
    def _setup_dataset_object(dataset_id, location=None):
        """
        Setup dataset object.
        """

        dataset = bigquery.Dataset(dataset_id)

        ## TODO: not being used since 1.6.0 - need to redo the description tha goes to bigquery
        dataset.description = "Para saber mais acesse https://basedosdados.org/"
        # dataset.description = self._render_template(
        #     Path("dataset/dataset_description.txt"), self.dataset_config
        # )

        dataset.location = location

        return dataset

    def _write_readme_file(self):
        """
        Write README.md file.
        """

        readme_content = (
            f"Como capturar os dados de {self.dataset_id}?\n\nPara cap"
            f"turar esses dados, basta verificar o link dos dados orig"
            f"inais indicado em dataset_config.yaml no item website.\n"
            f"\nCaso tenha sido utilizado algum código de captura ou t"
            f"ratamento, estes estarão contidos em code/. Se o dado pu"
            f"blicado for em sua versão bruta, não existirá a pasta co"
            f"de/.\n\nOs dados publicados estão disponíveis em: https:"
            f"//basedosdados.org/dataset/{self.dataset_id.replace('_','-')}"
        )

        readme_path = Path(self.metadata_path / self.dataset_id / "README.md")

        with open(readme_path, "w", encoding="utf-8") as readmefile:
            readmefile.write(readme_content)

    def init(self, replace=False):
        """Initialize dataset folder at metadata_path at `metadata_path/<dataset_id>`.

        The folder should contain:

        * `dataset_config.yaml`
        * `README.md`

        Args:
            replace (str): Optional. Whether to replace existing folder.

        Raises:
            FileExistsError: If dataset folder already exists and replace is False
        """

        # Create dataset folder
        try:
            self.dataset_folder.mkdir(exist_ok=replace, parents=True)
        except FileExistsError as e:
            raise FileExistsError(
                f"Dataset {str(self.dataset_folder.stem)} folder does not exists. "
                "Set replace=True to replace current files."
            ) from e

        # create dataset_config.yaml with metadata
        self.metadata.create(if_exists="replace")

        # create README.md file
        self._write_readme_file()

        # Add code folder
        (self.dataset_folder / "code").mkdir(exist_ok=replace, parents=True)

        return self

    def publicize(self, mode="all", dataset_is_public=True):
        """Changes IAM configuration to turn BigQuery dataset public.

        Args:
            mode (bool): Which dataset to create [prod|staging|all].
            dataset_is_public (bool): Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.
        """

        for m in self._loop_modes(mode):

            dataset = m["client"].get_dataset(m["id"])
            entries = dataset.access_entries
            # TODO https://github.com/basedosdados/mais/pull/1020
            # TODO if staging dataset is private, the prod view can't acess it: if dataset_is_public and "staging" not in dataset.dataset_id:
            if dataset_is_public:
                if "staging" not in dataset.dataset_id:
                    entries.extend(
                        [
                            bigquery.AccessEntry(
                                role="roles/bigquery.dataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                            bigquery.AccessEntry(
                                role="roles/bigquery.metadataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                            bigquery.AccessEntry(
                                role="roles/bigquery.user",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                        ]
                    )
                else:
                    entries.extend(
                        [
                            bigquery.AccessEntry(
                                role="roles/bigquery.dataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                        ]
                    )
                dataset.access_entries = entries
            m["client"].update_dataset(dataset, ["access_entries"])
        logger.success(
            " {object} {object_id}_{mode} was {action}!",
            object_id=self.dataset_id,
            mode=mode,
            object="Dataset",
            action="publicized",
        )

    def create(
        self, mode="all", if_exists="raise", dataset_is_public=True, location=None
    ):
        """Creates BigQuery datasets given `dataset_id`.

        It can create two datasets:

        * `<dataset_id>` (mode = 'prod')
        * `<dataset_id>_staging` (mode = 'staging')

        If `mode` is all, it creates both.

        Args:
            mode (str): Optional. Which dataset to create [prod|staging|all].
            if_exists (str): Optional. What to do if dataset exists

                * raise : Raises Conflict exception
                * replace : Drop all tables and replace dataset
                * update : Update dataset description
                * pass : Do nothing

            dataset_is_public (bool): Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.

            location (str): Optional. Location of dataset data.
                List of possible region names locations: https://cloud.google.com/bigquery/docs/locations


        Raises:
            Warning: Dataset already exists and if_exists is set to `raise`
        """

        if if_exists == "replace":
            self.delete(mode)
        elif if_exists == "update":

            self.update()
            return

        # Set dataset_id to the ID of the dataset to create.
        for m in self._loop_modes(mode):

            # Construct a full Dataset object to send to the API.
            dataset_obj = self._setup_dataset_object(m["id"], location=location)

            # Send the dataset to the API for creation, with an explicit timeout.
            # Raises google.api_core.exceptions.Conflict if the Dataset already
            # exists within the project.
            try:
                m["client"].create_dataset(dataset_obj)  # Make an API request.
                logger.success(
                    " {object} {object_id}_{mode} was {action}!",
                    object_id=self.dataset_id,
                    mode=mode,
                    object="Dataset",
                    action="created",
                )

            except Conflict as e:
                if if_exists == "pass":
                    return
                raise Conflict(f"Dataset {self.dataset_id} already exists") from e

        # Make prod dataset public
        self.publicize(dataset_is_public=dataset_is_public)

    def delete(self, mode="all"):
        """Deletes dataset in BigQuery. Toogle mode to choose which dataset to delete.

        Args:
            mode (str): Optional.  Which dataset to delete [prod|staging|all]
        """

        for m in self._loop_modes(mode):

            m["client"].delete_dataset(m["id"], delete_contents=True, not_found_ok=True)
        logger.info(
            " {object} {object_id}_{mode} was {action}!",
            object_id=self.dataset_id,
            mode=mode,
            object="Dataset",
            action="deleted",
        )

    def update(self, mode="all", location=None):
        """Update dataset description. Toogle mode to choose which dataset to update.

        Args:
            mode (str): Optional. Which dataset to update [prod|staging|all]
            location (str): Optional. Location of dataset data.
                List of possible region names locations: https://cloud.google.com/bigquery/docs/locations

        """

        for m in self._loop_modes(mode):

            # Send the dataset to the API to update, with an explicit timeout.
            # Raises google.api_core.exceptions.Conflict if the Dataset already
            # exists within the project.
            m["client"].update_dataset(
                self._setup_dataset_object(
                    m["id"],
                    location=location,
                ),
                fields=["description"],
            )  # Make an API request.

        logger.success(
            " {object} {object_id}_{mode} was {action}!",
            object_id=self.dataset_id,
            mode=mode,
            object="Dataset",
            action="updated",
        )



'''
Importing the module will automatically import the submodules.
'''


import sys
import os

__version__ = "1.6.11-b1"

sys.path.append(f"{os.getcwd()}/python-package")

#pylint: disable=C0413

from basedosdados.constants import constants, config
from basedosdados.upload.dataset import Dataset
from basedosdados.upload.storage import Storage
from basedosdados.upload.table import Table
from basedosdados.upload.metadata import Metadata
from basedosdados.download.base import reauth
from basedosdados.download.download import (
    read_sql,
    download,
    read_table,
)
from basedosdados.download.metadata import (
    list_datasets,
    list_dataset_tables,
    get_table_description,
    get_dataset_description,
    get_table_columns,
    get_table_size,
    search,
)



'''
Functions to get metadata from BD's API
'''
# pylint: disable=invalid-name,use-maxsplit-arg,line-too-long
from collections import defaultdict
import math

from google.cloud import bigquery
import pandas as pd
import requests

def _safe_fetch(url: str):
    """
    Safely fetchs urls and, if somehting goes wrong, informs user what is the possible cause
    """
    response = None
    try:
        response = requests.get(url, timeout=10)
        response.raise_for_status()
    except requests.exceptions.HTTPError as errh:
        print("Http Error:", errh)
    except requests.exceptions.ConnectionError as errc:
        print("Error Connecting:", errc)
    except requests.exceptions.Timeout as errt:
        print("Timeout Error:", errt)
    except requests.exceptions.RequestException as err:
        print("This url doesn't appear to exists:", err)

    return response


def _dict_from_page(json_response):
    """
    Generate a dict from BD's API response with dataset_id and description as keys
    """
    temp_dict = {
        "dataset_id": [
            dataset["name"] for dataset in json_response["result"]["datasets"]
        ],
        "description": [
            dataset["notes"] if "notes" in dataset.keys() else None
            for dataset in json_response["result"]["datasets"]
        ],
    }

    return temp_dict


def _fix_size(s, step=80):

    final = ""

    for l in s.split(" "):
        final += (l + " ") if len(final.split("\n")[-1]) < step else "\n"

    return final


def _print_output(df):
    """Prints dataframe contents as print blocks
    Args:
        df (pd.DataFrame): table to be printed
    """

    columns = df.columns
    step = 80
    print()
    for _, row in df.iterrows():
        for c in columns:
            print(_fix_size(f"{c}: \n\t{row[c]}"))
        print("-" * (step + 15))
    print()


def _handle_output(verbose, output_type, df, col_name=None):
    """Handles datasets and tables listing outputs based on user's choice.
    Either prints it to the screen or returns it as a `list` object.
    Args:
        verbose (bool): amount of verbosity
        output_type (str): type of output
        df (pd.DataFrame, bigquery.Dataset or bigquery.Table): table containing datasets metadata
        col_name (str): name of column with id's data
    """

    df_is_dataframe = isinstance(df,pd.DataFrame)
    df_is_bq_dataset_or_table = isinstance(df, bigquery.Table)
    df_is_bq_dataset_or_table |= isinstance(df, bigquery.Dataset)

    if verbose is True and df_is_dataframe:
        _print_output(df)

    elif verbose is True and df_is_bq_dataset_or_table:
        print(df.description)

    elif verbose is False:
        if output_type == "list":
            return df[col_name].to_list()
        if output_type == "str":
            return df.description
        if output_type == "records":
            return df.to_dict("records")
        msg = '`output_type` argument must be set to "list", "str" or "records".'
        raise ValueError(msg)
    raise TypeError("`verbose` argument must be of `bool` type.")

def list_datasets(with_description=False, verbose=True):
    """
    This function uses `bd_dataset_search` website API
    enpoint to retrieve a list of available datasets.

    Args:
        with_description (bool): Optional
            If True, fetch short dataset description for each dataset.
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, a list object is returned.

    Returns:
        list | stdout
    """
    # first request is made separately since we need to now the number of pages before the iteration
    page_size = 100  # this function will only made more than one requisition if there are more than 100 datasets in the API response #pylint: disable=C0301
    url = f"https://basedosdados.org/api/3/action/bd_dataset_search?q=&resource_type=bdm_table&page=1&page_size={page_size}"  # pylint: disable=C0301
    response = _safe_fetch(url)
    json_response = response.json()
    n_datasets = json_response["result"]["count"]
    n_pages = math.ceil(n_datasets / page_size)
    temp_dict = _dict_from_page(json_response)

    temp_dicts = [temp_dict]
    for page in range(2, n_pages + 1):
        url = f"https://basedosdados.org/api/3/action/bd_dataset_search?q=&resource_type=bdm_table&page={page}&page_size={page_size}"  # pylint: disable=C0301
        response = _safe_fetch(url)
        json_response = response.json()
        temp_dict = _dict_from_page(json_response)
        temp_dicts.append(temp_dict)

    dataset_dict = defaultdict(list)

    for d in temp_dicts:  # pylint: disable=C0103
        for key, value in d.items():
            dataset_dict[key].append(value)

    # flat inner lists
    dataset_dict["dataset_id"] = [
        item for sublist in dataset_dict["dataset_id"] for item in sublist
    ]  # pylint: disable=C0301
    dataset_dict["description"] = [
        item for sublist in dataset_dict["description"] for item in sublist
    ]  # pylint: disable=C0301
    # select desired output using dataset_id info. Note that the output is either a standardized string or a list #pylint: disable=C0301
    if verbose & (with_description is False):
        return _print_output(pd.DataFrame.from_dict(dataset_dict)[["dataset_id"]])
    if verbose & with_description:
        return _print_output(
            pd.DataFrame.from_dict(dataset_dict)[["dataset_id", "description"]]
        )
    if (verbose is False) & (with_description is False):
        return dataset_dict["dataset_id"]
    if (verbose is False) & with_description:
        return [
            {
                "dataset_id": dataset_dict["dataset_id"][k],
                "description": dataset_dict["description"][k],
            }
            for k in range(len(dataset_dict["dataset_id"]))
        ]
    raise ValueError("`verbose` and `with_description` argument must be of `bool` type.")


def list_dataset_tables(
    dataset_id,
    with_description=False,
    verbose=True,
):
    """
    Fetch table_id for tables available at the specified dataset_id. Prints the information on screen or returns it as a list.

    Args:
        dataset_id (str): Optional.
            Dataset id returned by list_datasets function
        limit (int):
            Field to limit the number of results
        with_description (bool): Optional
             If True, fetch short table descriptions for each table that match the search criteria.
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, a list object is returned.

    Returns:
        stdout | list
    """

    dataset_id = dataset_id.replace(
        "-", "_"
    )  # The dataset_id pattern in the bd_dataset_search endpoint response uses a hyphen as a separator, while in the endpoint urls that specify the dataset_id parameter the separator used is an underscore. See issue #1079

    url = f"https://basedosdados.org/api/3/action/bd_bdm_dataset_show?dataset_id={dataset_id}"

    response = _safe_fetch(url)

    json_response = response.json()

    dataset = json_response["result"]
    # this dict has all information need to output the function
    table_dict = {
        "table_id": [
            dataset["resources"][k]["name"]
            for k in range(len(dataset["resources"]))
            if dataset["resources"][k]["resource_type"] == "bdm_table"
        ],
        "description": [
            dataset["resources"][k]["description"]
            for k in range(len(dataset["resources"]))
            if dataset["resources"][k]["resource_type"] == "bdm_table"
        ],
    }
    # select desired output using table_id info. Note that the output is either a standardized string or a list
    if verbose & (with_description is False):
        return _print_output(pd.DataFrame.from_dict(table_dict)[["table_id"]])
    if verbose & with_description:
        return _print_output(
            pd.DataFrame.from_dict(table_dict)[["table_id", "description"]]
        )
    if (verbose is False) & (with_description is False):
        return table_dict["table_id"]
    if (verbose is False) & with_description:
        return [
            {
                "table_id": table_dict["table_id"][k],
                "description": table_dict["description"][k],
            }
            for k in range(len(table_dict["table_id"]))
        ]

    raise ValueError("`verbose` and `with_description` argument must be of `bool` type.")


def get_dataset_description(
    dataset_id,
    verbose=True,
):
    """
    Prints the full dataset description.

    Args:
        dataset_id (str): Required.
            Dataset id available in list_datasets.
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, data is returned as a `str`.

    Returns:
        stdout | str
    """
    url = f"https://basedosdados.org/api/3/action/bd_bdm_dataset_show?dataset_id={dataset_id}"

    response = _safe_fetch(url)

    json_response = response.json()

    description = json_response["result"]["notes"]

    if verbose:
        return print(description)
    return description


def get_table_description(
    dataset_id,
    table_id,
    verbose=True,
):
    """
    Prints the full table description.

    Args:
        dataset_id (str): Required.
            Dataset id available in list_datasets.
        table_id (str): Required.
            Table id available in list_dataset_tables
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, data is returned as a `str`.

    Returns:
        stdout | str
    """

    url = f"https://basedosdados.org/api/3/action/bd_bdm_table_show?dataset_id={dataset_id}&table_id={table_id}"

    response = _safe_fetch(url)

    json_response = response.json()

    description = json_response["result"]["description"]

    if verbose:
        return print(description)
    return description


def get_table_columns(
    dataset_id,
    table_id,
    verbose=True,
):

    """
        Fetch the names, types and descriptions for the columns in the specified table. Prints
        information on screen.
    Args:
        dataset_id (str): Required.
            Dataset id available in list_datasets.
        table_id (str): Required.
            Table id available in list_dataset_tables
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, data is returned as a `list` of `dict`s.

    Returns:
        stdout | list
    """

    url = f"https://basedosdados.org/api/3/action/bd_bdm_table_show?dataset_id={dataset_id}&table_id={table_id}"

    response = _safe_fetch(url)

    json_response = response.json()

    columns = json_response["result"]["columns"]

    if verbose:
        return _print_output(pd.DataFrame(columns))
    return columns


def get_table_size(
    dataset_id,
    table_id,
    verbose=True,
):
    """Use a query to get the number of rows and size (in Mb) of a table.

    WARNING: this query may cost a lot depending on the table.

    Args:
        dataset_id (str): Optional.
            Dataset id available in basedosdados. It should always come with table_id.
        table_id (str): Optional.
            Table id available in basedosdados.dataset_id.
            It should always come with dataset_id.
        verbose (bool): Optional.
            If set to True, information is printed to the screen. If set to False, data is returned as a `list` of `dict`s.
    """
    url = f"https://basedosdados.org/api/3/action/bd_bdm_table_show?dataset_id={dataset_id}&table_id={table_id}"

    response = _safe_fetch(url)

    json_response = response.json()

    size = json_response["result"]["size"]

    if size is None:
        return print("Size not available")
    if verbose:
        return _print_output(pd.DataFrame(size))
    return size


def search(query, order_by):
    """This function works as a wrapper to the `bd_dataset_search` website API
    enpoint.

    Args:
        query (str):
            String to search in datasets and tables' metadata.
        order_by (str): score|popular|recent
            Field by which the results will be ordered.

    Returns:
        pd.DataFrame:
            Response from the API presented as a pandas DataFrame. Each row is
            a table. Each column is a field identifying the table.
    """

    # validate order_by input
    if order_by not in ["score", "popular", "recent"]:
        raise ValueError(
            f'order_by must be score, popular or recent. Received "{order_by}"'
        )

    url = f"https://basedosdados.org/api/3/action/bd_dataset_search?q={query}&order_by={order_by}&resource_type=bdm_table"

    response = _safe_fetch(url)

    json_response = response.json()

    dataset_dfs = []
    # first loop identify the number of the tables in each datasets
    for dataset in json_response["result"]["datasets"]:
        tables_dfs = []
        len(dataset["resources"])
        # second loop extracts tables' information for each dataset
        for table in dataset["resources"]:
            data_table = pd.DataFrame(
                {k: str(table[k]) for k in list(table.keys())}, index=[0]
            )
            tables_dfs.append(data_table)
        # append tables' dataframes for each dataset
        data_ds = tables_dfs[0].append(tables_dfs[1:]).reset_index(drop=True)
        dataset_dfs.append(data_ds)
    # append datasets' dataframes
    df = dataset_dfs[0].append(dataset_dfs[1:]).reset_index(drop=True)

    return df



"""
Functions for managing downloads
"""
# pylint: disable=too-many-arguments, fixme, invalid-name, protected-access,line-too-long
from pathlib import Path
from functools import partialmethod
import re
import time
import shutil
import os
import gzip

from pydata_google_auth.exceptions import PyDataCredentialsError
from google.cloud import bigquery_storage_v1
from google.cloud import bigquery
import pandas_gbq
from pandas_gbq.gbq import GenericGBQException

from basedosdados.download.base import google_client, credentials
from basedosdados.exceptions import (
    BaseDosDadosException,
    BaseDosDadosAccessDeniedException,
    BaseDosDadosAuthorizationException,
    BaseDosDadosInvalidProjectIDException,
    BaseDosDadosNoBillingProjectIDException,
)
from basedosdados.constants import config


def _set_config_variables(billing_project_id, from_file):
    """
    Set billing_project_id and from_file variables
    """

    # standard billing_project_id configuration
    billing_project_id = billing_project_id or config.billing_project_id
    # standard from_file configuration
    from_file = from_file or config.from_file

    return billing_project_id, from_file


def read_sql(
    query,
    billing_project_id=None,
    from_file=False,
    reauth=False,
    use_bqstorage_api=False,
):
    """Load data from BigQuery using a query. Just a wrapper around pandas.read_gbq

    Args:
        query (sql):
            Valid SQL Standard Query to basedosdados
        billing_project_id (str): Optional.
            Project that will be billed. Find your Project ID here https://console.cloud.google.com/projectselector2/home/dashboard
        from_file (boolean): Optional.
            Uses the credentials from file, located in `~/.basedosdados/credentials/
        reauth (boolean): Optional.
            Re-authorize Google Cloud Project in case you need to change user or reset configurations.
        use_bqstorage_api (boolean): Optional.
            Use the BigQuery Storage API to download query results quickly, but at an increased cost(https://cloud.google.com/bigquery/docs/reference/storage/).
            To use this API, first enable it in the Cloud Console(https://console.cloud.google.com/apis/library/bigquerystorage.googleapis.com).
            You must also have the bigquery.readsessions.create permission on the project you are billing queries to.

    Returns:
        pd.DataFrame:
            Query result
    """

    billing_project_id, from_file = _set_config_variables(
        billing_project_id=billing_project_id, from_file=from_file
    )

    try:
        # Set a two hours timeout
        bigquery_storage_v1.client.BigQueryReadClient.read_rows = partialmethod(
            bigquery_storage_v1.client.BigQueryReadClient.read_rows,
            timeout=3600 * 2,
        )

        return pandas_gbq.read_gbq(
            query,
            credentials=credentials(from_file=from_file, reauth=reauth),
            project_id=billing_project_id,
            use_bqstorage_api=use_bqstorage_api,
        )

    except GenericGBQException as e:
        if "Reason: 403" in str(e):
            raise BaseDosDadosAccessDeniedException from e

        if re.match("Reason: 400 POST .* [Pp]roject[ ]*I[Dd]", str(e)):
            raise BaseDosDadosInvalidProjectIDException from e

        raise

    except PyDataCredentialsError as e:
        raise BaseDosDadosAuthorizationException from e

    except (OSError, ValueError) as e:
        no_billing_id = "Could not determine project ID" in str(e)
        no_billing_id |= "reading from stdin while output is captured" in str(e)
        if no_billing_id:
            raise BaseDosDadosNoBillingProjectIDException from e
        raise


def read_table(
    dataset_id,
    table_id,
    billing_project_id=None,
    query_project_id="basedosdados",
    limit=None,
    from_file=False,
    reauth=False,
    use_bqstorage_api=False,
):
    """Load data from BigQuery using dataset_id and table_id.

    Args:
        dataset_id (str): Optional.
            Dataset id available in basedosdados. It should always come with table_id.
        table_id (str): Optional.
            Table id available in basedosdados.dataset_id.
            It should always come with dataset_id.
        billing_project_id (str): Optional.
            Project that will be billed. Find your Project ID here https://console.cloud.google.com/projectselector2/home/dashboard
        query_project_id (str): Optional.
            Which project the table lives. You can change this you want to query different projects.
        limit (int): Optional.
            Number of rows to read from table.
        from_file (boolean): Optional.
            Uses the credentials from file, located in `~/.basedosdados/credentials/
        reauth (boolean): Optional.
            Re-authorize Google Cloud Project in case you need to change user or reset configurations.
        use_bqstorage_api (boolean): Optional.
            Use the BigQuery Storage API to download query results quickly, but at an increased cost(https://cloud.google.com/bigquery/docs/reference/storage/).
            To use this API, first enable it in the Cloud Console(https://console.cloud.google.com/apis/library/bigquerystorage.googleapis.com).
            You must also have the bigquery.readsessions.create permission on the project you are billing queries to.


    Returns:
        pd.DataFrame:
            Query result
    """

    billing_project_id, from_file = _set_config_variables(
        billing_project_id=billing_project_id, from_file=from_file
    )

    if (dataset_id is not None) and (table_id is not None):
        query = f"""
        SELECT * 
        FROM `{query_project_id}.{dataset_id}.{table_id}`"""

        if limit is not None:

            query += f" LIMIT {limit}"
    else:
        raise BaseDosDadosException("Both table_id and dataset_id should be filled.")

    return read_sql(
        query,
        billing_project_id=billing_project_id,
        from_file=from_file,
        reauth=reauth,
        use_bqstorage_api=use_bqstorage_api,
    )


def download(
    savepath,
    query=None,
    dataset_id=None,
    table_id=None,
    billing_project_id=None,
    query_project_id="basedosdados",
    limit=None,
    from_file=False,
    reauth=False,
    compression="GZIP",
):
    """Download table or query result from basedosdados BigQuery (or other).

    * Using a **query**:

        `download('select * from `basedosdados.br_suporte.diretorio_municipios` limit 10')`

    * Using **dataset_id & table_id**:

        `download(dataset_id='br_suporte', table_id='diretorio_municipios')`

    You can also add arguments to modify save parameters:

    `download(dataset_id='br_suporte', table_id='diretorio_municipios', index=False, sep='|')`


    Args:
        savepath (str, pathlib.PosixPath):
            savepath must be a file path. Only supports `.csv`.
        query (str): Optional.
            Valid SQL Standard Query to basedosdados. If query is available,
            dataset_id and table_id are not required.
        dataset_id (str): Optional.
            Dataset id available in basedosdados. It should always come with table_id.
        table_id (str): Optional.
            Table id available in basedosdados.dataset_id.
            It should always come with dataset_id.
        billing_project_id (str): Optional.
            Project that will be billed. Find your Project ID here https://console.cloud.google.com/projectselector2/home/dashboard
        query_project_id (str): Optional.
            Which project the table lives. You can change this you want to query different projects.
        limit (int): Optional
            Number of rows.
        from_file (boolean): Optional.
            Uses the credentials from file, located in `~/.basedosdados/credentials/
        reauth (boolean): Optional.
            Re-authorize Google Cloud Project in case you need to change user or reset configurations.
        compression (str): Optional.
            Compression type. Only `GZIP` is available for now.
    Raises:
        Exception: If either table_id, dataset_id or query are empty.
    """

    billing_project_id, from_file = _set_config_variables(
        billing_project_id=billing_project_id, from_file=from_file
    )

    if (query is None) and ((table_id is None) or (dataset_id is None)):
        raise BaseDosDadosException(
            "Either table_id, dataset_id or query should be filled."
        )

    client = google_client(billing_project_id, from_file, reauth)

    # makes sure that savepath is a filepath and not a folder
    savepath = _sets_savepath(savepath)

    # if query is not defined (so it won't be overwritten) and if
    # table is a view or external or if limit is specified,
    # convert it to a query.
    if not query and (
        not _is_table(client, dataset_id, table_id, query_project_id) or limit
    ):
        query = f"""
        SELECT * 
          FROM {query_project_id}.{dataset_id}.{table_id}
        """

        if limit is not None:
            query += f" limit {limit}"

    if query:
        # sql queries produces anonymous tables, whose names
        # can be found within `job._properties`
        job = client["bigquery"].query(query)

        # views may take longer: wait for job to finish.
        _wait_for(job)

        dest_table = job._properties["configuration"]["query"]["destinationTable"]

        project_id = dest_table["projectId"]
        dataset_id = dest_table["datasetId"]
        table_id = dest_table["tableId"]

    _direct_download(client, dataset_id, table_id, savepath, project_id, compression)


def _direct_download(
    client,
    dataset_id,
    table_id,
    savepath,
    project_id="basedosdados",
    compression="GZIP",
    extract=True,
):
    """
    Download file to disk without the requirement of loading it in memory.
    Creates a temporary file based on the `table_id` and the time of
    execution. Also create a temporary bucket using the `dataset_id`
    and the time of execution. Move the table to the temporary file and
    download it to disk. In the end, remove the temporary bucket.
    Args:
        client (dict of google.cloud.bigquery.client.Client):
            BigQuery and Storage clients.
        dataset_id (str):
            Dataset id available in project_id.
        table_id (str):
            Table id available in project_id.dataset_id.
        savepath (str, `pathlib.PosixPath`):
            Local path in which file should be stored in disk.
        project_id (str, optional):
            In case you want to use to query another project, by default 'basedosdados'
        compression (str, optional):
            Compression type to use for exported files.
            Can be one of ["NONE"|"GZIP"]. Defaults to GZIP.
        extract (bool, optional):
            Whether to extract the gzip file.
    """
    time_hash = str(hash(time.time()))

    # Bucket names must start and end with a number or letter.
    tmp_file_name = table_id
    tmp_bucket_name = _clean_name(dataset_id + "_" + time_hash)
    blob_path = f"gs://{tmp_bucket_name}/{tmp_file_name}-*"

    # Creates temporary savepath
    tmp_savepath = savepath.parent / "tmp"
    tmp_savepath.mkdir(parents=True, exist_ok=True)

    try:
        # create temporary bucket
        _create_bucket(client, tmp_bucket_name)

        # move table to temporary file inside temporary bucket
        _move_table_to_bucket(
            client, dataset_id, table_id, blob_path, project_id, compression
        )

        # download file from bucket directly to disk
        _download_blob_from_bucket(client, tmp_bucket_name, tmp_savepath)

        if compression == "GZIP" and extract:
            _gzip_extract(tmp_savepath)

        _join_files(tmp_savepath, savepath)

    except Exception as err:
        # TODO handle exceptions for 404 (not found), 403 (forbidden)
        raise Exception(err) from err
    finally:
        # delete temporary bucket (even in the case of crashing)
        _delete_bucket(client, tmp_bucket_name)
        # delete temporary savepath
        shutil.rmtree(tmp_savepath)


def _download_blob_from_bucket(client, bucket_name, savepath):
    """
    Download a blob from a bucket to the path specified.
    Args:
        client (dict of google.cloud.bigquery.client.Client):
            BigQuery and Storage clients.
        bucket_name (str):
            Name of the bucket for the file to be stored.
        blob_name (str):
            Name of the file to be downloaded from bucket.
        savepath (str):
            Local path in which file should be stored in disk.
    """

    bucket = client["storage"].bucket(bucket_name)
    for blob in bucket.list_blobs():
        filepath = savepath / (blob.name.split("-")[-1] + ".csv.gz")
        blob.download_to_filename(filepath)


def _create_bucket(client, bucket_name):
    """
    Create a new buket in a specific location with standard storage class
    Args:
        client (dict of google.cloud.bigquery.client.Client):
            BigQuery and Storage clients.
        bucket_name (str):
            Name of the bucket to be created.
    """
    storage_client = client["storage"]
    bucket = storage_client.bucket(bucket_name)

    # standard storage class are adequate for data
    # stored for only brief periods of time
    bucket.storage_class = "STANDARD"

    storage_client.create_bucket(bucket, location="US")


def _delete_bucket(client, bucket_name):
    """
    Forceably deletes a bucket.
    This method deletes all blobs from a bucket.
    Args:
        client (dict of google.cloud.bigquery.client.Client):
            BigQuery and Storage clients.
        bucket_name (str):
            Name of the bucket to be deleted.
    """
    MAX_BLOBS = 256

    storage_client = client["storage"]
    bucket = storage_client.get_bucket(bucket_name)

    # NOTE: force=True implementation will not delete
    # the temporary bucket if n_blobs >= 256
    n_blobs = len(list(storage_client.list_blobs(bucket_name)))
    if n_blobs >= MAX_BLOBS:
        raise Exception(
            f"""Your temporary bucket contains more than {MAX_BLOBS}. You should manually delete it (force=True will not be able to delete it.)."""
        )

    bucket.delete(force=True)


def _move_table_to_bucket(
    client,
    dataset_id,
    table_id,
    blob_path,
    project_id="basedosdados",
    compression="GZIP",
):
    """
    Move table from BigQuery to bucket
    Args:
        client (dict of google.cloud.bigquery.client.Client):
            BigQuery and Storage clients.
        dataset_id (str):
            Dataset id available in project_id.
        table_id (str):
            Table id available in project_id.dataset_id.
        blob_path (str):
            Path in bucket where the table will be stored.
            Called ``destination_uri`` in the documentation in the form of
            ``gs://<bucket_name>/<file_name.csv>``
        project_id (str, optional):
            In case you want to use to query another
            project, defaults to 'basedosdados'
        compression (str, optional):
            Compression type to use for exported files.
            Can be one of ["NONE"|"GZIP"]
    """
    client = client["bigquery"]
    dataset_ref = bigquery.DatasetReference(project_id, dataset_id)
    table_ref = dataset_ref.table(table_id)

    job_config = bigquery.job.ExtractJobConfig(compression=compression)

    # perform transfer from bq to bucket
    extract_job = client.extract_table(
        table_ref,
        blob_path,
        location="US",
        job_config=job_config,
    )
    # wait for API results
    extract_job.result()


def _gzip_extract(savepath):
    """Extracts and replace gzip file"""
    try:
        for file in savepath.glob("*"):
            with gzip.open(file, "rb") as f_in:
                with open(file.with_suffix(""), "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)

            os.remove(file)
    except Exception as e:
        raise Exception("GZIP file could not be extracted.") from e


def _join_files(tmp_savepath, savepath):
    """Joins all files in savepath"""

    files = list(tmp_savepath.glob("*.csv"))
    with savepath.open("a+") as targetfile:
        for i, file in enumerate(files):
            with file.open("r") as f:
                if i > 0:
                    next(f)
                for line in f:
                    targetfile.write(line)

            os.remove(file)


def _is_table(client, dataset_id, table_id, project_id):
    """Check whether a `table_id` is a view or not."""

    if (dataset_id is None) or (table_id is None):
        return False

    dataset_ref = bigquery.DatasetReference(project_id, dataset_id)
    table_ref = dataset_ref.table(table_id)

    table = client["bigquery"].get_table(table_ref)

    return table.table_type == "TABLE"


def _clean_name(string):
    """Remove anything not a number or letter"""

    pattern = r"[^a-zA-Z0-9]+"
    replace = ""
    return re.sub(pattern, replace, string)


def _wait_for(job):
    """Wait for bigquery job to finish.
    Args:
        job (`bigquery.job`): bigquery job object from a
        ``bigquery.Client().query()`` call.
    """
    while not job.done():
        time.sleep(1)


def _sets_savepath(savepath):
    """Sets savepath accordingly"""

    savepath = Path(savepath)

    if savepath.suffix == ".csv":
        # make sure that path exists
        savepath.parent.mkdir(parents=True, exist_ok=True)
    else:
        raise BaseDosDadosException(
            "Only .csv files are supported, "
            f"your filename has a diferent extension: {savepath.suffix}"
        )

    return savepath



'''
Constants for the project.
'''
# pylint: disable=C0103
__all__ = ["config", "constants"]

from enum import Enum
from dataclasses import dataclass


@dataclass
class config:
    '''
    Configuration for the project.
    '''
    verbose: bool = True
    billing_project_id: str = None
    project_config_path: str = None
    from_file: bool = False


class constants(Enum):
    '''
    Constants for the project.
    '''
    ENV_CONFIG: str = "BASEDOSDADOS_CONFIG"
    ENV_CREDENTIALS_PREFIX: str = "BASEDOSDADOS_CREDENTIALS_"
    ENV_CREDENTIALS_PROD: str = "BASEDOSDADOS_CREDENTIALS_PROD"
    ENV_CREDENTIALS_STAGING: str = "BASEDOSDADOS_CREDENTIALS_STAGING"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['basedosdados',
 'basedosdados.cli',
 'basedosdados.download',
 'basedosdados.upload']

package_data = \
{'': ['*'],
 'basedosdados': ['configs/*',
                  'configs/templates/dataset/*',
                  'configs/templates/table/*']}

install_requires = \
['Jinja2==3.0.3',
 'ckanapi==4.6',
 'click==8.0.3',
 'google-cloud-bigquery-storage==1.1.0',
 'google-cloud-bigquery==2.30.1',
 'google-cloud-storage==1.42.3',
 'importlib-metadata>=4.11.3,<5.0.0',
 'loguru>=0.6.0,<0.7.0',
 'pandas-gbq>=0.17.4,<0.18.0',
 'pandas>=1.3.5,<2.0.0',
 'pandavro>=1.6.0,<2.0.0',
 'pyaml==20.4.0',
 'pyarrow==6.0.0',
 'ruamel.yaml==0.17.10',
 'shapely>=1.6.0,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'tomlkit==0.7.0',
 'tqdm==4.50.2']

entry_points = \
{'console_scripts': ['basedosdados = basedosdados.cli.cli:cli']}

setup_kwargs = {
    'name': 'basedosdados',
    'version': '1.6.11b3',
    'description': 'Organizar e facilitar o acesso a dados brasileiros através de tabelas públicas no BigQuery.',
    'long_description': '# Python Package\n\n## Desenvolvimento Linux e Mac: \n\nClone o repositório principal:\n\n```sh\ngit clone https://github.com/basedosdados/mais.git\n```\nEntre na pasta local do repositório usando `cd mais/` e suba o ambiente localmente:\n\n```sh\nmake create-env\n. .mais/bin/activate\ncd python-package/\npython setup.py develop\n```\n\n### Desenvolva uma nova feature\n\n1. Abra uma branch com o nome issue-<X>\n2. Faça as modificações necessárias\n3. Suba o Pull Request apontando para a branch `python-next-minor` ou `python-next-patch`. \n    Sendo, minor e patch referentes ao bump da versão: v1.5.7 --> v\\<major>.\\<minor>.\\<patch>.\n4. O nome do PR deve seguir o padrão\n    `[infra] <titulo explicativo>`\n\n\n### O que uma modificação precisa ter\n\n  \n- Resolver o problema\n- Lista de modificações efetuadas\n    1. Mudei a função X para fazer Y\n    2. Troquei o nome da variavel Z\n- Referência aos issues atendidos\n- Documentação e Docstrings\n- Testes\n  \n\n## Versionamento\n\n**Para publicar uma nova versão do pacote é preciso seguir os seguintes passos:**\n\n1. Fazer o pull da branch:\n\n    ```bash\n    git pull origin [python-version]\n    ```\n  \n    Onde `[python-version]` é a branch da nova versão do pacote.\n\n2. Se necessario adicionar novas dependências:\n    ```bash\n      poetry add <package-name>\n    ```\n\n3. Gerar novo `requirements-dev.txt` \n\n    ```bash\n    poetry export -f requirements.txt --output requirements-dev.txt --without-hashes\n    ```\n\n4. Editar `pyproject.toml`:\n\n    O arquivo `pyproject.toml` contém, entre outras informações, a versão do pacote em python da **BD**. Segue excerto do arquivo:\n\n    ```toml\n    description = "Organizar e facilitar o acesso a dados brasileiros através de tabelas públicas no BigQuery."\n    homepage = "https://github.com/base-dos-dados/bases"\n    license = "MIT"\n    name = "basedosdados"\n    packages = [\n      {include = "basedosdados"},\n    ]\n    readme = "README.md"\n    repository = "https://github.com/base-dos-dados/bases"\n    version = "1.6.1-beta.2"\n    ```\n    \n    O campo `version` deve ser alterado para o número da versão sendo lançada.\n\n5. Push para branch:\n\n    ```bash\n    git push origin [python-version]\n    ```\n\n6. Publicação do pacote no PyPI (exige usuário e senha):\n\n    Para publicar o pacote no PyPI, use:\n    \n    ```bash\n    poetry version [python-version]\n    poetry publish --build\n    ```\n\n7. Faz merge da branch para a master\n8. Faz release usando a UI do GitHub\n9. Atualizar versão do pacote usada internamente\n',
    'author': 'Joao Carabetta',
    'author_email': 'joao.carabetta@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/base-dos-dados/bases',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""
Exception classes for the package.
"""

#pylint: disable=C0301

class BaseDosDadosException(Exception):
    """Exclusive Exception from Base dos Dados"""


class BaseDosDadosAccessDeniedException(BaseDosDadosException):
    """Exception raised if the user provides a wrong GCP project ID."""

    def __init__(self):
        self.message = (
            "\nAre you sure you are using the right `billing_project_id`?"
            "\nYou must use the Project ID available in your Google Cloud"
            " console home page at https://console.cloud.google.com/home/dashboard"
            "\nIf you still don't have a Google Cloud Project, you have to "
            "create one.\n"
            "You can set one up by following these steps: \n"
            "1. Go to this link https://console.cloud.google.com/projectselector2/home/dashboard\n"
            "2. Agree with Terms of Service if asked\n"
            "3. Click in Create Project\n"
            "4. Put a cool name in your project\n"
            "5. Hit create\n"
            "6. Rerun this command with the flag `reauth=True`. \n"
            "   Like `read_table('br_ibge_pib', 'municipios', "
            "billing_project_id=<YOUR_PROJECT_ID>, reauth=True)`"
        )
        super().__init__(self.message)


class BaseDosDadosInvalidProjectIDException(BaseDosDadosException):
    """Exception raised if the user provides an invalid GCP project ID."""

    def __init__(self):
        self.message = (
            "\nYou are using an invalid `billing_project_id`.\nMake sure "
            "you set it to the Project ID available in your Google Cloud"
            " console home page at https://console.cloud.google.com/home/dashboard"
        )
        super().__init__(self.message)


class BaseDosDadosNoBillingProjectIDException(BaseDosDadosException):
    """Exception raised if the user provides no GCP billing project ID."""

    def __init__(self):
        self.message = (
            "\nWe are not sure which Google Cloud project should be billed.\n"
            "First, you should make sure that you have a Google Cloud project.\n"
            "If you don't have one, set one up following these steps: \n"
            "\t1. Go to this link https://console.cloud.google.com/projectselector2/home/dashboard\n"
            "\t2. Agree with Terms of Service if asked\n"
            "\t3. Click in Create Project\n"
            "\t4. Put a cool name in your project\n"
            "\t5. Hit create\n"
            "\n"
            "Copy the Project ID, (notice that it is not the Project Name)\n"
            "Now, you have two options:\n"
            "1. Add an argument to your function poiting to the billing project id.\n"
            "   Like `bd.read_table('br_ibge_pib', 'municipios', billing_project_id=<YOUR_PROJECT_ID>)`\n"
            "2. You can set a project_id in the environment by running the following command in your terminal: `gcloud config set project <YOUR_PROJECT_ID>`.\n"
            "   Bear in mind that you need `gcloud` installed."
        )
        super().__init__(self.message)


class BaseDosDadosAuthorizationException(BaseDosDadosException):
    """Exception raised if the user doesn't complete the authorization
    process correctly."""

    def __init__(self):
        self.message = (
            "\nAre you sure you did the authorization process correctly?\n"
            "If you were given the option to enter an authorization code, "
            "please try again and make sure you are entering the right one."
            "\nYou can try again by rerunning this command with the flag "
            "`reauth=True`. \n\tLike `read_table('br_ibge_pib', 'municipios',"
            " billing_project_id=<YOUR_PROJECT_ID>, reauth=True)`"
            "\nThen you can click at the provided link and get the right "
            "authorization code."
        )
        super().__init__(self.message)
